import json
from typing import List, Dict, Any

from yid_langchain_extensions.output_parser.stop_seq_output_parser import StopSeqOutputParser
from yid_langchain_extensions.output_parser.utils import escape_new_lines_in_json_values, close_all_curly_brackets, \
    strip_json_from_md_snippet


class FixingJSONParser(StopSeqOutputParser):
    @property
    def stop_sequences(self) -> List[str]:
        return ["}\n```", "}```"]

    @staticmethod
    def fix_json_md_snippet(text: str) -> str:
        fixed_json = text
        if fixed_json.startswith("'") or fixed_json.startswith('"'):
            fixed_json = fixed_json[1:]
        fixed_json = fixed_json.strip()
        fixed_json = escape_new_lines_in_json_values(fixed_json)
        fixed_json = close_all_curly_brackets(fixed_json)
        if "```json" in fixed_json:
            fixed_json = fixed_json[fixed_json.find("```json"):]
        else:
            fixed_json = "```json\n" + fixed_json
        if not fixed_json.endswith("```"):
            fixed_json += "\n```"
        return fixed_json

    @staticmethod
    def parse_json_md_snippet(json_md_snippet: str) -> Dict[str, Any]:
        cleaned_json = strip_json_from_md_snippet(json_md_snippet)
        response = json.loads(cleaned_json)
        return response

    def parse(self, text: str) -> Dict[str, Any]:
        fixed_json_md_snippet = self.fix_json_md_snippet(text)
        return self.parse_json_md_snippet(fixed_json_md_snippet)



from typing import List, Dict
from unittest import TestCase

from langchain_community.llms.fake import FakeListLLM
from langchain_core.language_models import LLM
from langchain_core.messages import BaseMessage, AIMessage, HumanMessage, SystemMessage
from langchain_core.runnables import RunnableSerializable
from langchain_openai import ChatOpenAI
from pydantic import BaseModel

from yid_langchain_extensions.utils import pydantic_v1_port, FirstMessageAuthorContextSizeLimiter, \
    NaiveContextSizeLimiter


class Pydantic2ClassWithoutPort(BaseModel):
    llm: LLM


class Pydantic2ClassWithPort(BaseModel):
    llm: pydantic_v1_port(LLM)


class Pydantic2ClassWithPortNested(BaseModel):
    llm_list: pydantic_v1_port(List[LLM])
    llm_dict: pydantic_v1_port(Dict[str, LLM])
    complex_field: pydantic_v1_port(RunnableSerializable[List[BaseMessage], str])


class TestPydanticV1Port(TestCase):
    def setUp(self):
        self.llm = FakeListLLM(responses=[])

    def test_v1_port(self):
        Pydantic2ClassWithPort(llm=self.llm)

    def test_port_required(self):
        with self.assertRaises(TypeError):
            Pydantic2ClassWithoutPort(llm=self.llm)

    def test_v1_port_nested(self):
        Pydantic2ClassWithPortNested(
            llm_list=[self.llm],
            llm_dict={"llm": self.llm},
            complex_field=self.llm
        )


class TestFirstMessageAuthorContextSizeLimiter(TestCase):
    def test_limit_messages(self):
        messages = [
            HumanMessage(content="hi"),
            SystemMessage(content="say hi"),
            AIMessage(content="hi"),
            HumanMessage(content="how r u?"),
            SystemMessage(content="say how r u"),
            AIMessage(content="im fine"),
        ]
        limiter = FirstMessageAuthorContextSizeLimiter(
            first_message_author="human",
            base_limiter=NaiveContextSizeLimiter(
                max_context_size=40, llm=ChatOpenAI()
            )
        )
        limited_messages = limiter.limit_messages(messages)
        self.assertEqual(len(limited_messages), 3)



from typing import List

from langchain_core.prompts import PromptTemplate
from pydantic.v1 import BaseModel, validator

from yid_langchain_extensions.output_parser.fixing_json_parser import FixingJSONParser

FORMAT_PROMPT = """RESPONSE FORMAT:
------
You response should be a markdown code snippet formatted in the following schema:

```json
{
    {{thoughts}}
}
```"""


class Thought(BaseModel):
    name: str
    description: str
    type: str = "string"


class ThoughtsJSONParser(FixingJSONParser):
    thoughts: List[Thought]
    format_prompt: str = FORMAT_PROMPT

    @validator("thoughts")
    def validate_thoughts(cls, v):  # noqa
        assert len(v) > 0, "You must have at least one thought"
        return v

    def format_thoughts(self) -> str:
        return "\n\t".join([
            f'"{thought.name}": {thought.type} [{thought.description}]' for thought in self.thoughts
        ])

    def get_format_instructions(self, **kwargs) -> str:
        """
        :param kwargs: If your thoughts or format_prompt have some extra placeholders, you can fill them by kwargs
        """
        format_instructions = PromptTemplate.from_template(self.format_prompt, template_format="jinja2").partial(
            thoughts=self.format_thoughts()).format_prompt(**kwargs).to_string()
        return format_instructions



from langchain.schema import BaseOutputParser


class DirectParser(BaseOutputParser):
    def parse(self, text: str) -> str:
        return text

    def get_format_instructions(self) -> str:
        return ""



import unittest

from langchain import hub
from langchain.agents.output_parsers.openai_tools import OpenAIToolsAgentOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI
from pydantic import BaseModel as BaseModelV2, Field as FieldV2

from yid_langchain_extensions.llm.tools_llm_with_thought import build_tools_llm_with_thought
from yid_langchain_extensions.output_parser.pydantic_from_tool import PydanticOutputParser
from yid_langchain_extensions.utils import convert_to_openai_tool_v2


class PowerFnArgsV2(BaseModelV2):
    power: int = FieldV2(description="power")
    base: float = FieldV2(description="base")


class Reasoning(BaseModelV2):
    reasoning: str


class TestToolsLLMWithThought(unittest.TestCase):
    def test(self):
        tools = [
            convert_to_openai_tool_v2(PowerFnArgsV2)
        ]
        base_prompt = hub.pull("hwchase17/openai-tools-agent")
        thought_prompt: ChatPromptTemplate = hub.pull("dimitree54/introduce_thought_tool")
        llm = ChatOpenAI(model="gpt-3.5-turbo-0125", temperature=0)  # noqa
        llm_with_thought = build_tools_llm_with_thought(llm, tools, thought_prompt, Reasoning)
        extended_chain = base_prompt | llm_with_thought | PydanticOutputParser(
            pydantic_class=PowerFnArgsV2, base_parser=OpenAIToolsAgentOutputParser())
        result = extended_chain.invoke({"input": "3.43^5", "agent_scratchpad": [], "chat_history": []})
        self.assertEqual(result.base, 3.43)
        self.assertEqual(result.power, 5)



import base64
from abc import ABC, abstractmethod
from typing import Annotated, Type, Any, Dict, Union, Callable, Optional, Sequence, List

import cv2
import numpy as np
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_core.prompt_values import ChatPromptValue
from langchain_core.runnables import RunnableConfig, Runnable
from langchain_core.tools import BaseTool
from langchain_core.utils.function_calling import _rm_titles, convert_to_openai_function, convert_to_openai_tool  # noqa
from langchain_core.utils.json_schema import dereference_refs
from pydantic import ValidationError, PlainValidator, BaseModel as BaseModelV2, BaseModel
from pydantic.v1 import BaseModel as BaseModelV1, parse_obj_as


def validated(value: BaseModelV1, target_class: Type[Any]) -> BaseModelV1:
    try:
        return parse_obj_as(target_class, value)  # noqa
    except ValidationError as e:
        raise ValidationError(f"Input of invalid type: {e}")


def pydantic_v1_port(cls):
    return Annotated[cls, PlainValidator(lambda val: validated(val, cls))]


def convert_pydantic_to_openai_function_v2(
        model: Type[BaseModelV2],
) -> Dict[str, Any]:
    schema = dereference_refs(model.model_json_schema())
    schema.pop("definitions", None)
    title = schema.pop("title", "")
    default_description = schema.pop("description", "")
    return {
        "name": title,
        "description": default_description,
        "parameters": _rm_titles(schema)
    }


def convert_to_openai_function_v2(
        function: Union[Dict[str, Any], Type[BaseModelV1], Callable, BaseTool, Type[BaseModelV2]],
) -> Dict[str, Any]:
    if isinstance(function, type) and issubclass(function, BaseModelV2):
        return convert_pydantic_to_openai_function_v2(function)
    else:
        return convert_to_openai_function(function)


def convert_to_openai_tool_v2(
        tool: Union[Dict[str, Any], Type[BaseModelV1], Callable, BaseTool, Type[BaseModelV2]],
) -> Dict[str, Any]:
    if isinstance(tool, type) and issubclass(tool, BaseModelV2):
        return convert_to_openai_tool(
            convert_to_openai_function_v2(tool)
        )
    else:
        return convert_to_openai_tool(tool)


class ChatPromptValue2DictAdapter(Runnable[Union[ChatPromptValue, Dict], Dict[str, Sequence[BaseMessage]]]):
    def invoke(
            self, input: Union[ChatPromptValue, Dict], config: Optional[RunnableConfig] = None  # noqa
    ) -> Dict[str, Sequence[BaseMessage]]:
        if isinstance(input, Dict):
            return input
        return {
            "messages": input.messages
        }


class ContextSizeLimiter(BaseModel, ABC):
    @abstractmethod
    def limit_messages(self, messages: List[BaseMessage]) -> List[BaseMessage]:
        pass


class NaiveContextSizeLimiter(ContextSizeLimiter):
    max_context_size: int
    llm: pydantic_v1_port(BaseChatModel)

    def limit_messages(self, messages: List[BaseMessage]) -> List[BaseMessage]:
        num_messages_to_keep = 1
        while num_messages_to_keep <= len(messages):
            extended_num_messages_to_keep = num_messages_to_keep + 1
            last_messages = messages[-extended_num_messages_to_keep:]
            tokens_in_extended_messages = self.llm.get_num_tokens_from_messages(last_messages)
            if tokens_in_extended_messages > self.max_context_size:
                break
            num_messages_to_keep = extended_num_messages_to_keep
        return messages[-num_messages_to_keep:]


class FirstMessageAuthorContextSizeLimiter(ContextSizeLimiter):
    first_message_author: str
    base_limiter: ContextSizeLimiter

    def _remove_head_until_author(self, messages: List[BaseMessage]) -> List[BaseMessage]:
        for i in range(len(messages)):
            if messages[i].type == self.first_message_author:
                return messages[i:]
        return []

    def limit_messages(self, messages: List[BaseMessage]) -> List[BaseMessage]:
        messages = self.base_limiter.limit_messages(messages)
        return self._remove_head_until_author(messages)


def encode_image_to_url(image: np.ndarray) -> str:
    # Ensure the image is in 3-channel format
    if len(image.shape) != 3 or image.shape[2] != 3:
        raise ValueError("Input image must be a 3-channel image")

    # Encode the image to PNG format
    _, buffer = cv2.imencode('.png', image)

    # Convert the buffer to base64
    base64_image = base64.b64encode(buffer).decode('utf-8')

    # Create the data URL
    base64_url = f"data:image/png;base64,{base64_image}"

    return base64_url



from typing import List

from pydantic import BaseModel

from yid_langchain_extensions.output_parser.thoughts_json_parser import Thought, ThoughtsJSONParser


class Class(BaseModel):
    name: str
    description: str


def get_classes_description(classes: List[Class]) -> str:
    result = ""
    for class_index, _class in enumerate(classes):
        result += f"> {class_index}: {_class.name}. {_class.description};\n"
    return result


def get_classes_summary(classes: List[Class]) -> str:
    return "; ".join([f"{class_index} ({_class.name})" for class_index, _class in enumerate(classes)])


class ClassParser(ThoughtsJSONParser):
    @classmethod
    def from_extra_thoughts(cls, extra_thoughts: List[Thought]):
        action_thoughts: List[Thought] = [
            Thought(name="class_index", type="int",
                    description="The class chosen. Must be one of [{{classes_summary}}]"),
        ]
        thoughts = extra_thoughts + action_thoughts
        return cls(thoughts=thoughts)

    def parse(self, text: str) -> int:
        response = super().parse(text)
        return int(response["class_index"])



from typing import Optional, Tuple

from langchain.agents import AgentExecutor
from langchain.schema import AgentFinish, AgentAction

from yid_langchain_extensions.output_parser.action_parser import ActionWithThoughts


class AgentWithThoughtsExecutor(AgentExecutor):
    def _get_tool_return(
        self, next_step_output: Tuple[AgentAction, str]
    ) -> Optional[AgentFinish]:
        agent_action, observation = next_step_output
        name_to_tool_map = {tool.name: tool for tool in self.tools}
        if agent_action.tool in name_to_tool_map and name_to_tool_map[agent_action.tool].return_direct:
            return_values = {self.agent.return_values[0]: observation}
            if isinstance(agent_action, ActionWithThoughts):
                return_values.update(agent_action.all_thoughts)
            return AgentFinish(
                return_values,
                agent_action.log
            )
        return None



import json
from typing import Dict, List


def escape_new_lines_in_json_values(json_string: str) -> str:
    result = ""
    i_am_inside_quotes = False
    for v in json_string:
        if v == '"':
            i_am_inside_quotes = not i_am_inside_quotes
        if v == "\n" and i_am_inside_quotes:
            result += "\\n"
        else:
            result += v
    return result


def close_all_curly_brackets(json_string: str) -> str:
    result = json_string
    num_brackets_to_close = json_string.count("{") - json_string.count("}")
    result += "\n}" * num_brackets_to_close
    return result


def strip_json_from_md_snippet(json_md_snippet: str) -> str:
    cleaned_text = json_md_snippet[len("```json"): -len("```")]
    return cleaned_text


def get_dict_without_extra_fields(dict_to_strip: Dict, keys_to_keep: List[str]) -> Dict:
    stripped_dict = {}
    for key, value in dict_to_strip.items():
        if key in keys_to_keep:
            stripped_dict[key] = value
    return stripped_dict


def format_dict_to_json_md(dict_to_format: Dict) -> str:
    return "```json\n" + json.dumps(dict_to_format, indent=4) + "\n```"
